# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""SKOS Import code specific to cubicweb >= 3.21"""

from itertools import imap

from six import text_type

from cubicweb.dataimport.importer import (ExtEntity, ExtEntitiesImporter, RelationMapping,
                                          cwuri2eid, use_extid_as_cwuri)
from cubicweb.dataimport.importer import HTMLImportLog  # noqa

from cubes.skos.sobjects import ext_dump_relations


class SKOSRelationMapping(RelationMapping):

    def __getitem__(self, rtype):
        if rtype in ('exact_match', 'close_match'):
            rql = 'Any S,O WHERE S {0} O'.format(rtype)
            return set(tuple(x) for x in self.cnx.execute(rql))
        return super(SKOSRelationMapping, self).__getitem__(rtype)


def store_skos_extentities(cnx, store, entities, import_log,
                           source=None, raise_on_error=False,
                           extid2eid=None, extid_as_cwuri=True):
    """Add SKOS external entities to the store. Don't commit/flush any data.

    Arguments:

    * `cnx`, RQL connection to the CubicWeb instance
    * `store`, store to use for the import
    * `entities`, iterable (usualy a generator) on external entities to import
    * `import_log`, import log instance to give to the store
    * `source`, optional source in which existing entities will be looked for
      (default to the system source)
    * `raise_on_error`, boolean flag telling if we should fail on error or simply log it (default
      to `False`)
    * `extid_as_cwuri`, boolean flag telling if we should use the external entities'extid as `cwuri`
      attribute of imported entities (default to `True`)
    """
    # only consider the system source for schemes and labels
    if source is None:
        source_eid = cnx.repo.system_source.eid
    else:
        source_eid = source.eid
    if extid2eid is None:
        extid2eid = cwuri2eid(cnx, ('ConceptScheme', 'Label'), source_eid=source_eid)
        # though concepts and external URIs may come from any source
        extid2eid.update(cwuri2eid(cnx, ('Concept', 'ExternalUri')))
    # plug function that turn previously known external uris by newly inserted concepts
    restore_relations = {}

    # pylint: disable=dangerous-default-value
    def externaluri_to_concept(extentity, cnx=cnx, extid2eid=extid2eid,
                               restore_relations=restore_relations):
        try:
            eid = extid2eid[extentity.extid]
        except KeyError:
            pass
        else:
            if extentity.etype == 'Concept' and cnx.entity_metas(eid)['type'] == 'ExternalUri':
                # We have replaced the external uri by the new concept. As entities.extid column is
                # unique, we've to drop the external uri before inserting the concept, so we:
                #  1. record every relations from/to the external uri,
                #  2. remove it,
                #  3. insert the concept and
                #  4. reinsert relations using the concept instead
                #
                # 1. record relations from/to the external uri
                restore_relations[extentity.extid] = ext_dump_relations(cnx, extid2eid, extentity)
                # 2. remove the external uri entity
                cnx.execute('DELETE ExternalUri X WHERE X eid %(x)s', {'x': eid})
                # 3. drop its extid from the mapping to trigger insertion of the concept by the
                # importer
                del extid2eid[extentity.extid]
                # 4. will be done in SKOSExtEntitiesImporter
        return extentity

    entities = imap(externaluri_to_concept, entities)
    # plug function to detect the concept scheme
    concept_schemes = []

    def record_scheme(extentity):
        if extentity.etype == 'ConceptScheme':
            concept_schemes.append(extentity.extid)
        return extentity

    entities = imap(record_scheme, entities)
    etypes_order_hint = ('ConceptScheme', 'Concept', 'Label')
    existing_relations = SKOSRelationMapping(cnx, source)
    importer = SKOSExtEntitiesImporter(schema=cnx.vreg.schema, store=store,
                                       extid2eid=extid2eid, existing_relations=existing_relations,
                                       restore_relations=restore_relations,
                                       etypes_order_hint=etypes_order_hint, import_log=import_log,
                                       raise_on_error=raise_on_error)
    if extid_as_cwuri:
        set_cwuri = use_extid_as_cwuri(importer.extid2eid)
        entities = set_cwuri(entities)
    importer.import_entities(entities)
    stats = (importer.created, importer.updated)
    return stats, concept_schemes


class SKOSExtEntitiesImporter(ExtEntitiesImporter):
    """Override ExtEntitiesImporter to handle creation of additionnal relations to newly created
    concepts that replace a former external uri, and to create ExternalUri entities for URIs used in
    exact_match / close_match relations which have no known entity in the repository yet.
    """

    def __init__(self, *args, **kwargs):
        self.restore_relations = kwargs.pop('restore_relations')
        super(SKOSExtEntitiesImporter, self).__init__(*args, **kwargs)

    def prepare_insert_entity(self, extentity):
        eid = super(SKOSExtEntitiesImporter, self).prepare_insert_entity(extentity)
        # (4.) restore relations formerly from/to an equivalent external uri
        try:
            relations = self.restore_relations.pop(extentity.extid)
        except KeyError:
            return eid
        for subject_eid, rtype, object_eid in relations:
            if subject_eid is None:
                subject_eid = eid
            if object_eid is None:
                object_eid = eid
            self.store.prepare_insert_relation(subject_eid, rtype, object_eid)
        return eid

    def prepare_insert_deferred_relations(self, deferred):
        # create missing targets for exact_match and close_match relations
        for rtype in ('exact_match', 'close_match'):
            relations = deferred.get(rtype, ())
            for _, object_uri in relations:
                if object_uri not in self.extid2eid:
                    extentity = ExtEntity('ExternalUri', object_uri,
                                          values=dict(uri=text_type(object_uri),
                                                      cwuri=text_type(object_uri)))
                    self.prepare_insert_entity(extentity)
        return super(SKOSExtEntitiesImporter, self).prepare_insert_deferred_relations(deferred)



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import sys
from io import StringIO

from cubicweb.devtools import testlib
from cubicweb.cwconfig import CubicWebConfiguration
try:
    from cubicweb.devtools import PostgresApptestConfiguration, startpgcluster, stoppgcluster
except ImportError as exc:
    import pytest
    pytestmark = pytest.mark.skipif(True, reason=str(exc))
else:
    from cubes.skos import ccplugin


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


class ImportSkosDataCommandTC(testlib.CubicWebTC):
    configcls = PostgresApptestConfiguration

    def setup_database(self):
        super(ImportSkosDataCommandTC, self).setup_database()
        self.orig_config_for = CubicWebConfiguration.config_for
        config_for = lambda appid: self.config
        CubicWebConfiguration.config_for = staticmethod(config_for)

    def tearDown(self):
        CubicWebConfiguration.config_for = self.orig_config_for
        super(ImportSkosDataCommandTC, self).tearDown()

    def run_import_skos(self, fpath, *args, **kwargs):
        cmd = [self.appid, fpath] + list(args)
        sys.stdout = StringIO()
        try:
            ccplugin.ImportSkosData(None).main_run(cmd)
        finally:
            stdout_check = kwargs.pop('stdout_check', None)
            if stdout_check is not None:
                self.assertIn(stdout_check, sys.stdout.getvalue())
            sys.stdout = sys.__stdout__

    def _test_base(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.find('ConceptScheme').one()
            self.assertEqual(scheme.title, u"Thésaurus de test")
            concept = cnx.find('Concept', cwuri='http://mystery.com/ark:/kra/543').one()
            expected = {
                u'fr': u'économie',
                u'fr-fr': u'économie',
            }
            self.assertEqual(concept.labels, expected)

    def test_nooption(self):
        self.run_import_skos(self.datapath('skos.rdf'))
        self._test_base()

    def test_nohook(self):
        self.run_import_skos(self.datapath('skos.rdf'),
                             '--cw-store', 'nohook')
        self._test_base()

    def test_massive(self):
        self.run_import_skos(self.datapath('skos.rdf'),
                             '-s', 'massive')
        self._test_base()

    def test_lcsv(self):
        self.run_import_skos(self.datapath('lcsv_example_shortened.csv'),
                             '--format', 'lcsv')
        with self.admin_access.cnx() as cnx:
            rset = cnx.find('ConceptScheme', title=u'lcsv_example_shortened.csv')
            self.assertTrue(rset)
            scheme = rset.one()
            self.assertEqual(scheme.title, u'lcsv_example_shortened.csv')
            self.assertEqual(len(scheme.reverse_in_scheme), 5)

    def test_lcsv_rollback(self):
        self.run_import_skos(
            self.datapath('lcsv_example_missing_prolog.csv'),
            '--format', 'lcsv',
            stdout_check=u"Aborting due to: 'missing prolog column (#)'",
        )
        with self.admin_access.cnx() as cnx:
            rset = cnx.find('ConceptScheme',
                            title=u'lcsv_example_missing_prolog.csv')
        self.assertFalse(rset)


if __name__ == '__main__':
    from unittest import main
    main()



# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-ctl plugin introducing skos-import command.
"""
from __future__ import print_function

import logging
from os import path

from cubicweb.toolsutils import Command, underline_title
from cubicweb.cwctl import CWCTL
from cubicweb.utils import admincnx
from cubicweb.dataimport.importer import SimpleImportLog

from cubes.skos import rdfio
from cubes.skos.sobjects import graph_extentities, import_skos_extentities, lcsv_extentities


def _massive_store_factory(cnx, **kwargs):
    from cubicweb.dataimport.massive_store import MassiveObjectStore
    return MassiveObjectStore(cnx, **kwargs)


def _nohook_store_factory(cnx, **kwargs):
    from cubicweb.dataimport.stores import NoHookRQLObjectStore
    return NoHookRQLObjectStore(cnx, **kwargs)


def _rql_store_factory(cnx, **kwargs):
    from cubicweb.dataimport.stores import RQLObjectStore
    return RQLObjectStore(cnx, **kwargs)


class LoggingImportLog(SimpleImportLog):
    logger = logging.getLogger('skos.import')

    def __init__(self):
        pass

    def _log(self, severity, msg, path, line):
        self.logger.log(severity, msg)


class ImportSkosData(Command):
    """Import a RDF SKOS concept scheme.

    <instance>
      identifier of the instance into which the scheme will be imported. Should use the skos cube.

    <filepath>
      path to the SKOS files to import.

    """
    arguments = '[options] <instance> <filepath>...'
    name = 'skos-import'
    min_args = 2
    options = (
        ('format',
         {'type': 'choice',
          'choices': ('rdf', 'lcsv'),
          'default': 'rdf',
          'help': 'format of input data (rdf or lcsv)'
          }),
        ('cw-store',
         {'short': 's',
          'type': 'choice', 'choices': ('rql', 'nohook', 'massive'), 'default': 'rql',
          'help': 'cubicweb store type: rql, nohook or massive (from the most secure but slowest '
          'to the less secure but fastest).'
          }),
        ('rdf-store',
         {'short': 'r',
          'type': 'choice', 'choices': ('librdf', 'rdflib'), 'default': 'rdflib',
          'help': 'RDF store type: librdf or rdflib (only with --format rdf).'
          }),
    )

    rdf_store_factories = {
        'librdf': rdfio.LibRDFRDFGraph,
        'rdflib': rdfio.RDFLibRDFGraph,
    }

    cw_store_factories = {
        'massive': _massive_store_factory,
        'nohook': _nohook_store_factory,
        'rql': _rql_store_factory,
    }

    def run(self, args):
        print(u'\n%s' % underline_title('Importing Skos dataset'))
        appid = args[0]
        connection = admincnx(appid)

        drop_scheme = None

        if self.get('format') == 'lcsv':
            try:
                fpath, = args[1:]
            except TypeError:
                raise Exception('LCSV format expects exactly one input file')
            with connection as cnx:
                title = path.basename(fpath).decode('utf-8')
                scheme = cnx.create_entity('ConceptScheme', title=title)
                cnx.commit()
                scheme_uri = scheme.cwuri

            def drop_scheme(cnx):
                """Drop ConceptScheme created prior to LCSV import."""
                rset = cnx.find('ConceptScheme', title=title)
                if rset:
                    rset.one().cw_delete()

            def extentities_generator():
                """ExtEntity generator function holding control on `fpath` file."""
                with open(fpath) as stream:
                    for extentity in lcsv_extentities(stream, scheme_uri,
                                                      language_code=u'fr'):
                        yield extentity

            extentities = extentities_generator()
        else:
            graph = self.rdf_store_factories[self.get('rdf-store')]()
            for filepath in args[1:]:
                print(u'loading {} into RDF graph'.format(filepath))
                graph.load(filepath)
            extentities = graph_extentities(graph)

        import_log = LoggingImportLog()
        with connection as cnx:
            store = self.cw_store_factories[self.get('cw-store')](cnx)
            try:
                (created, updated), conceptschemes = import_skos_extentities(
                    cnx, extentities, import_log, store=store)
                cnx.commit()
            except BaseException as exc:
                cnx.rollback()
                if drop_scheme is not None:
                    drop_scheme(cnx)
                    cnx.commit()
                print(u"Aborting due to: '%s'" % exc)
            else:
                print(u'Created: %d\nUpdated: %d' % (len(created), len(updated)))
                if conceptschemes:
                    print(u'Concept schemes:\n* %s' % '\n* '.join(conceptschemes))
        cnx.repo.shutdown()


CWCTL.register(ImportSkosData)



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos site customizations"""

# pylint: disable=wrong-import-order

from datetime import datetime

from six import text_type

from logilab.common.decorators import monkeypatch

from cubes.skos import POST_321, POST_323

if not POST_321:
    from copy import copy
    import inspect

    from cubicweb.dataimport import MetaGenerator, NoHookRQLObjectStore

    @monkeypatch(MetaGenerator)
    def __init__(self, cnx, baseurl=None, source=None):
        self._cnx = cnx
        if baseurl is None:
            config = cnx.vreg.config
            baseurl = config['base-url'] or config.default_base_url()
        if not baseurl[-1] == '/':
            baseurl += '/'
        self.baseurl = baseurl
        if source is None:
            source = cnx.repo.system_source
        self.source = source
        self.create_eid = cnx.repo.system_source.create_eid
        self.time = datetime.now()
        # attributes/relations shared by all entities of the same type
        self.etype_attrs = []
        self.etype_rels = []
        # attributes/relations specific to each entity
        self.entity_attrs = ['cwuri']
        schema = cnx.vreg.schema
        rschema = schema.rschema
        for rtype in self.META_RELATIONS:
            # skip owned_by / created_by if user is the internal manager
            if cnx.user.eid == -1 and rtype in ('owned_by', 'created_by'):
                continue
            if rschema(rtype).final:
                self.etype_attrs.append(rtype)
            else:
                self.etype_rels.append(rtype)

    @monkeypatch(MetaGenerator)
    def init_entity(self, entity):
        entity.eid = self.create_eid(self._cnx)
        extid = entity.cw_edited.get('cwuri')
        for attr in self.entity_attrs:
            if attr in entity.cw_edited:
                # already set, skip this attribute
                continue
            genfunc = self.generate(attr)
            if genfunc:
                entity.cw_edited.edited_attribute(attr, genfunc(entity))
        if isinstance(extid, text_type):
            extid = extid.encode('utf-8')
        return self.source, extid

    @monkeypatch(NoHookRQLObjectStore)
    def create_entity(self, etype, **kwargs):
        for key, value in kwargs.iteritems():
            kwargs[key] = getattr(value, 'eid', value)
        entity, rels = self.metagen.base_etype_dicts(etype)
        # make a copy to keep cached entity pristine
        entity = copy(entity)
        entity.cw_edited = copy(entity.cw_edited)
        entity.cw_clear_relation_cache()
        entity.cw_edited.update(kwargs, skipsec=False)
        entity_source, extid = self.metagen.init_entity(entity)
        cnx = self._cnx
        self.source.add_info(cnx, entity, entity_source, extid)
        self.source.add_entity(cnx, entity)
        kwargs = dict()
        if inspect.getargspec(self.add_relation).keywords:
            kwargs['subjtype'] = entity.cw_etype
        for rtype, targeteids in rels.iteritems():
            # targeteids may be a single eid or a list of eids
            inlined = self.rschema(rtype).inlined
            try:
                for targeteid in targeteids:
                    self.add_relation(cnx, entity.eid, rtype, targeteid,
                                      inlined, **kwargs)
            except TypeError:
                self.add_relation(cnx, entity.eid, rtype, targeteids,
                                  inlined, **kwargs)
        self._nb_inserted_entities += 1
        return entity


if not POST_323:
    # asynchronous source synchronization from the UI (https://www.cubicweb.org/ticket/10468967)
    # other part in views/__init__.py

    from pytz import utc
    from cubicweb import ObjectNotFound
    from cubicweb.server.sources import datafeed

    @monkeypatch(datafeed.DataFeedSource)
    def pull_data(self, cnx, force=False, raise_on_error=False, import_log_eid=None):
        """Launch synchronization of the source if needed.

        This method is responsible to handle commit/rollback on the given
        connection.
        """
        if not force and self.fresh():
            return {}
        if not self.acquire_synchronization_lock(cnx, force):
            return {}
        try:
            return self._pull_data(cnx, force, raise_on_error, import_log_eid)
        finally:
            cnx.rollback()  # rollback first in case there is some dirty transaction remaining
            self.release_synchronization_lock(cnx)

    @monkeypatch(datafeed.DataFeedSource)
    def _pull_data(self, cnx, force=False, raise_on_error=False, import_log_eid=None):
        importlog = self.init_import_log(cnx, import_log_eid)
        myuris = self.source_cwuris(cnx)
        try:
            parser = self._get_parser(cnx, sourceuris=myuris, import_log=importlog)
        except ObjectNotFound:
            return {}
        if self.process_urls(parser, self.urls, raise_on_error):
            self.warning("some error occurred, don't attempt to delete entities")
        else:
            parser.handle_deletion(self.config, cnx, myuris)
        self.update_latest_retrieval(cnx)
        stats = parser.stats
        if stats.get('created'):
            importlog.record_info('added %s entities' % len(stats['created']))
        if stats.get('updated'):
            importlog.record_info('updated %s entities' % len(stats['updated']))
        importlog.write_log(cnx, end_timestamp=self.latest_retrieval)
        cnx.commit()
        return stats

    @monkeypatch(datafeed.DataFeedSource)
    def init_import_log(self, cnx, import_log_eid=None, **kwargs):
        if import_log_eid is None:
            import_log = cnx.create_entity('CWDataImport', cw_import_of=self,
                                           start_timestamp=datetime.now(tz=utc),
                                           **kwargs)
        else:
            import_log = cnx.entity_from_eid(import_log_eid)
            import_log.cw_set(start_timestamp=datetime.now(tz=utc), **kwargs)
        cnx.commit()
        import_log.init()
        return import_log

    @monkeypatch(datafeed.DataFeedSource)
    def acquire_synchronization_lock(self, cnx, force=False):
        # XXX race condition until WHERE of SET queries is executed using
        # 'SELECT FOR UPDATE'
        now = datetime.now(tz=utc)
        if force:
            maxdt = now
        else:
            maxdt = now - self.max_lock_lifetime
        if not cnx.execute(
                'SET X in_synchronization %(now)s WHERE X eid %(x)s, '
                'X in_synchronization NULL OR X in_synchronization < %(maxdt)s',
                {'x': self.eid, 'now': now, 'maxdt': maxdt}):
            self.error('concurrent synchronization detected, skip pull')
            cnx.commit()
            return False
        cnx.commit()
        return True




from yams.buildobjs import RelationDefinition, EntityType
from cubicweb.schema import RQLConstraint


class inlined_match(RelationDefinition):
    subject = ('Concept', 'ExternalUri')
    object = ('Concept', 'ExternalUri')
    cardinality = '?*'
    inlined = True


class scheme_relation_type(RelationDefinition):
    subject = 'ConceptScheme'
    object = 'CWRType'
    cardinality = '**'


class Language(EntityType):
    u""""""


class language_to(RelationDefinition):
    subject = 'Language'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation_type CR, '
                                 'CR name "language_to"')]



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""SKOS import is available through:

* a c-c command
* a repository service
* the datafeed source API

The first two will create local entities (i.e. whose `cw_source` relation points to system source)
while the last will create entities belonging to the external source.

This file contains generic import code, the datafeed parser and the repository service.
"""

from hashlib import md5
from uuid import uuid4
from itertools import imap

from logilab.common.deprecation import deprecated

from cubicweb import schema, dataimport
from cubicweb.predicates import match_kwargs, match_user_groups
from cubicweb.server import Service
from cubicweb.server.sources import datafeed

from cubes.skos import (POST_321, LABELS_RDF_MAPPING, register_skos_rdf_input_mapping,
                        rdfio, lcsv, to_unicode)


def ext_dump_relations(cnx, extid2eid, extentity):
    """Wrapper around :func:`dump_relations` but receiving extentity and extdid2eid mapping as
    argument, to attempt to work around store's limitation on inlined relation.
    """
    eid = extid2eid[extentity.extid]
    etype = cnx.entity_metas(eid)['type']
    relations = []
    rschema = cnx.vreg.schema.rschema
    for subj, rtype, obj in dump_relations(cnx, eid, etype):
        if rschema(rtype).inlined:
            if subj is None:
                if obj is None:
                    obj = extentity.extid
                else:
                    extid2eid[obj] = obj
                extentity.values.setdefault(rtype, set()).add(obj)
            else:
                raise Exception("can't dump inlined object relation {0}".format(rtype))
        else:
            relations.append((subj, rtype, obj))
    return relations


def dump_relations(cnx, eid, etype):
    """Return a list of relation 3-uples `(subject_eid, relation, object_eid)` with None instead of
    `subject_eid` or `object_eid` depending on whether the entity type corresponding to `eid` is
    subject or object.
    """
    eschema = cnx.vreg.schema.eschema(etype)
    relations = []
    for rschema, _, role in eschema.relation_definitions():
        if rschema.rule:  # computed relation
            continue
        rtype = rschema.type
        if rtype in schema.VIRTUAL_RTYPES or rtype in schema.META_RTYPES:
            continue
        if role == 'subject':
            for object_eid, in cnx.execute('Any Y WHERE X %s Y, X eid %%(x)s' % rtype, {'x': eid}):
                if object_eid == eid:
                    object_eid = None
                relations.append((None, rtype, object_eid))
        else:
            for subject_eid, in cnx.execute('Any Y WHERE Y %s X, X eid %%(x)s' % rtype, {'x': eid}):
                if subject_eid == eid:
                    subject_eid = None
                relations.append((subject_eid, rtype, None))
    return relations


class RDFSKOSImportService(Service):
    """CubicWeb service to import a ConceptScheme from SKOS RDF."""

    __regid__ = 'rdf.skos.import'
    __select__ = match_kwargs('stream') & match_user_groups('managers')

    def call(self, stream, **kwargs):
        import_log = HTMLImportLog(getattr(stream, 'filename', u''))
        (created, updated), scheme_uris = self._do_import(stream, import_log, **kwargs)
        import_log.record_info('%s entities created' % len(created))
        import_log.record_info('%s entities updated' % len(updated))
        self._cw.commit()
        return import_log.logs, scheme_uris

    # Extracted method to let a chance to client cubes to give more arguments to
    # import_skos_extentities function
    def _do_import(self, stream, import_log, rdf_format=None, **kwargs):
        entities = rdf_extentities(stream, rdf_format)
        return import_skos_extentities(self._cw, entities, import_log, **kwargs)


class LCSVSKOSImportService(RDFSKOSImportService):
    """CubicWeb service to import a Concept and Label entities from LCSV file into a pre-existing
    ConceptScheme.
    """
    __regid__ = 'lcsv.skos.import'

    # Extracted method to let a chance to client cubes to give more arguments to
    # import_skos_extentities function
    def _do_import(self, stream, import_log,
                   scheme_uri, delimiter, encoding, language_code=None,
                   **kwargs):
        entities = lcsv_extentities(stream, scheme_uri, delimiter, encoding, language_code)
        return import_skos_extentities(self._cw, entities, import_log,
                                       extid_as_cwuri=False, **kwargs)


class SKOSParser(datafeed.DataFeedParser):
    """CubicWeb parser to import a ConceptScheme from SKOS RDF."""
    __regid__ = 'rdf.skos'

    def process(self, url, raise_on_error=False):
        """Main entry point"""
        try:
            created, updated = self._do_import(url, raise_on_error=raise_on_error)[0]
        except Exception as ex:
            if raise_on_error:
                raise
            self.exception('error while importing %s', url)
            self.import_log.record_error(to_unicode(ex))
            return True
        self.stats['created'] = created
        self.stats['updated'] = updated
        return False

    # Extracted method to let a chance to client cubes to give more arguments to
    # import_skos_extentities function
    def _do_import(self, url, raise_on_error, **kwargs):
        try:
            rdf_format = rdfio.guess_rdf_format(url)
        except ValueError:
            rdf_format = 'xml'
        entities = rdf_extentities(url, rdf_format)
        return import_skos_extentities(self._cw, entities, self.import_log,
                                       source=self.source, raise_on_error=raise_on_error, **kwargs)


def lcsv_extentities(stream, scheme_uri, delimiter=None, encoding='utf-8', language_code=None):
    """Return external entities generator from SKOS LCSV stream or URL (by transforming it to RDF
    first).

    Arguments:

    * `stream`, stream of the LCSV data
    * `scheme_uri`, cwuri of  the scheme into which imported concepts will be inserted
    * `delimiter`, CSV delimiter character
    * `encoding`, encoding of the CSV file
    * `language_code`, default language code of labels, in case it's not specified along the label
      definition
    """
    graph = rdfio.RDFLibRDFGraph()
    # add LCSV statements to the RDF graph
    lcsv2rdf = lcsv.LCSV2RDF(stream, delimiter=delimiter, encoding=encoding,
                             default_lang=language_code, uri_cls=graph.uri,
                             uri_generator=lambda x: str(uuid4()) + x)
    for (subj, pred, obj) in lcsv2rdf.triples():
        graph.add(subj, pred, obj)

    # we need an extra transform that link Concept to the parent concept skeme
    def relate_concepts(extentity):
        """Relate Concept ExtEntities to the ConceptScheme"""
        if extentity.etype == 'Concept':
            extentity.values.setdefault('in_scheme', set([])).add(scheme_uri)
        return extentity

    return imap(relate_concepts, graph_extentities(graph))


def rdf_extentities(stream_or_url, rdf_format=None):
    """Return external entities generator from SKOS RDF stream or URL.

    Arguments:

    * `stream_or_url`, stream or URL of the RDF data

    * `rdf_format`, RDF format used to defined the data. See formats supported by
      :class:`RDFLibRDFGraph`. If not specified, it will be guessed from the URL extension or
      default to xml
    """
    graph = rdfio.RDFLibRDFGraph()
    graph.load(stream_or_url, rdf_format)
    return graph_extentities(graph)


def graph_extentities(graph):
    """Return external entities generator from a RDF graph implementation (i.e. instance
    implementing :class:`cubes.skos.rdfio.AbstractRDFGraph`).
    """
    reg = rdfio.RDFRegistry()
    register_skos_rdf_input_mapping(reg)
    all_label_rtypes = frozenset(LABELS_RDF_MAPPING)
    # transform direct mapping of RDF as ExtEntity into Yams compatible ExtEntity by turning label
    # literals into entities.
    for extentity in rdfio.rdf_graph_to_entities(reg, graph, ('ConceptScheme', 'Concept')):
        label_rtypes = frozenset(extentity.values) & all_label_rtypes
        if not label_rtypes:
            yield extentity
            continue
        labels = []
        for rtype in label_rtypes:
            kind = rtype.split('_', 1)[0]  # drop '_label' suffix
            for label in extentity.values.pop(rtype):
                lang = getattr(label, 'lang', None)
                md5hash = md5(str(lang) + label.encode('utf-8'))
                labelid = str(extentity.extid) + '#' + rtype + md5hash.hexdigest()
                labels.append(ExtEntity('Label', labelid,
                                        {'label': set([label]),
                                         'language_code': set([lang]),
                                         'kind': set([kind]),
                                         'label_of': set([extentity.extid])}))
        # yield extentity before labels since it must be handled first in case the ExternalUri
        # to Concept transformation apply
        yield extentity
        for label in labels:
            yield label


def import_skos_extentities(cnx, entities, import_log,
                            source=None, metagenerator=None, store=None, **kwargs):
    """Import SKOS external entities.

    Arguments:

    * `cnx`, RQL connection to the CubicWeb instance
    * `entities`, iterable (usualy a generator) on external entities to import
    * `import_log`, import log instance to give to the store
    * `source`, optional source in which existing entities will be looked for
      (default to the system source); if store and metagenerator are
      unspecified, imported entities will be attached to this source,
      otherwise make sure the source declared in store/metagenerator is
      consistent with this one
    * `metagenerator`, metadata generator to be given to the store if store isn't specified
    * `store`, store to use for the import

    Extra arguments will be given to :func:`store_skos_extentities` function.
    """
    # By default, use NoHookRQLObjectStore for insertion in the database, with a custom
    # metagenerator to properly set the source for created entities
    if store is None:
        if metagenerator is None:
            metagenerator = dataimport.MetaGenerator(cnx, source=source)
        store = dataimport.NoHookRQLObjectStore(cnx, metagenerator)
    else:
        assert metagenerator is None, 'you should not give both a store and a metadata generator'
    res = store_skos_extentities(cnx, store, entities, import_log, source, **kwargs)
    store.flush()
    store.commit()
    if hasattr(store, 'finish'):  # XXX not all store have this before cw 3.22
        store.finish()
    return res


if POST_321:
    # use new CW 3.21 dataimport API
    from .post321_import import HTMLImportLog, ExtEntity, store_skos_extentities
else:
    from .pre321_import import HTMLImportLog, ExtEntity, store_skos_extentities


@deprecated('[skos 0.9] use lcsv_extentities then import_skos_extentities instead')
def _skos_import_lcsv(cnx, stream, import_log, scheme_uri, delimiter,
                      encoding, language_code=None, **kwargs):
    """Import SKOS from LCSV stream or URL (by transforming it to RDF first)."""
    entities = lcsv_extentities(stream, scheme_uri, delimiter, encoding, language_code)
    return import_skos_extentities(cnx, entities, import_log, extid_as_cwuri=False, **kwargs)


@deprecated('[skos 0.9] use rdf_extentities then import_skos_extentities instead')
def _skos_import_rdf(cnx, stream_or_url, import_log, rdf_format=None, **kwargs):
    """Import SKOS from RDF stream or URL."""
    entities = rdf_extentities(stream_or_url, rdf_format)
    return import_skos_extentities(cnx, entities, import_log, **kwargs)



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from cubicweb import tags
from cubicweb.predicates import is_instance

try:
    from cubes.relationwidget import views as rwdg
except ImportError:
    pass
else:

    class SearchForRelatedConceptsView(rwdg.SearchForRelatedEntitiesView):
        """Specializes relation widget to link to concepts which are expected to be constrained by
        an RQL expression on the edited relation definition.

        It will attempt to select the best suited language for display an return properly
        prefetched/ordered concepts. It's also expected that all concepts will provide a preferred
        label for the selected language (among those defined in the scheme).
        """
        __select__ = (rwdg.SearchForRelatedEntitiesView.__select__
                      & rwdg.edited_relation(tetype='Concept'))

        has_creation_form = False

        def linkable_rset(self):
            """Return rset of entities to be displayed as possible values for the
            edited relation. You may want to override this.
            """
            baserql, args = self.constrained_rql()
            # detect language first
            languages = set(lang for lang, in self._cw.execute(
                'DISTINCT Any OLC WHERE ' + baserql + ', O preferred_label OL, '
                'OL language_code OLC', args))
            for lang in self.preferred_languages():
                if lang in languages:
                    break
            else:
                if not languages:
                    return self._cw.empty_rset()
                lang = languages.pop()
            # now build the query to prefetch/sort on the preferred label in the proper language
            args['lang'] = lang
            return self._cw.execute(
                'DISTINCT Any O,OL,OLL ORDERBY OLL WHERE ' + baserql + ', O preferred_label OL, '
                'OL label OLL, OL language_code %(lang)s', args)

        def constrained_rql(self):
            constraints = self.schema_rdef.constraints
            assert len(constraints) == 1
            baserql = constraints[0].expression
            return baserql, {}

        @property
        def schema_rdef(self):
            rtype, tetype, role = self.rdef
            entity = self.compute_entity()
            return entity.e_schema.rdef(rtype, role, tetype)

        def preferred_languages(self):
            """Return language by order of preference."""
            return (self._cw.lang, 'en', None)

        @staticmethod
        def pref_label_label(concept):
            return concept.cw_rset[concept.cw_row][2]

        @staticmethod
        def pref_label_column(w, concept):
            w(tags.a(SearchForRelatedConceptsView.pref_label_label(concept),
                     href=concept.absolute_url()))

    class SelectConceptEntitiesTableView(rwdg.SelectEntitiesTableView):
        """Table view of the selectable entities in the relation widget

        Selection of columns (and respective renderer) can be overridden by
        updating `columns` and `column_renderers` class attributes.
        """
        __select__ = (rwdg.SelectEntitiesTableView.__select__ & is_instance('Concept'))

        column_renderers = rwdg.SelectEntitiesTableView.column_renderers.copy()
        # speed-up things by considering rset shape from ContainedSearchForRelatedEntitiesView
        column_renderers['entity'] = rwdg.SelectMainEntityColRenderer(
            sortfunc=SearchForRelatedConceptsView.pref_label_label,
            renderfunc=SearchForRelatedConceptsView.pref_label_column)



# copyright 2015-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Utilities for RDF import/export"""

from os.path import abspath, splitext
from urlparse import urlparse

from six import PY2, text_type, string_types

from cubes.skos import POST_321
if POST_321:
    # use new CW 3.21 dataimport API
    from cubicweb.dataimport.importer import ExtEntity
else:
    from cubes.skos import ExtEntity


TYPE_PREDICATE_URI = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'
SAME_AS_PREDICATE_URI = 'http://www.w3.org/2002/07/owl#sameAs'


class unicode_with_language(text_type):  # pylint: disable=invalid-name
    """Extend an unicode string to hold a .lang attribute as well"""

    def __new__(cls, content, lang):
        self = text_type.__new__(cls, content)
        self.lang = lang
        return self

    def __repr__(self):
        return '<%r(%s)>' % (super(unicode_with_language, self).__repr__(), self.lang)

    def __eq__(self, other):
        if isinstance(other, unicode_with_language):
            return (text_type(self), self.lang) == (text_type(other), other.lang)
        return False

    def __ne__(self, other):
        return not self == other


def normalize_uri(uri, prefixes, raise_on_error=False):
    """Normalize uri by attempting to substitute prefix by the associated namespace. Returns the
    normalized URI.
    """
    try:
        prefix, value = uri.split(':', 1)
    except ValueError:
        return uri
    try:
        return prefixes[prefix] + value
    except KeyError:
        if raise_on_error:
            raise RDFRegistryError('prefix {0} not found'.format(prefix))
        return uri


def permanent_url(entity):
    """Return permanent URL for the given entity (`<site url>/<eid>`)."""
    return entity._cw.build_url(text_type(entity.eid))


EXTENSION_FORMAT_MAP = {
    '.xml': 'xml',
    '.rdf': 'xml',
    '.rdfs': 'xml',
    '.owl': 'xml',
    '.n3': 'n3',
    '.ttl': 'n3',
    '.nt': 'nt',
    '.trix': 'trix',
    '.xhtml': 'rdfa',
    '.html': 'rdfa',
    '.svg': 'rdfa',
    '.nq': 'nquads',
    '.trig': 'trig'
}


def guess_rdf_format(filename):
    """Return RDF format as detected by the file extension, or raise ValueError if it can't be
    detected.

    Mapping between extension and format is defined in `EXTENSION_FORMAT_MAP` and formats known by
    default are 'xml', 'n3', 'nt', 'trix', 'rdfa', 'nquads'.
    """
    ext = splitext(filename)[1].lower()
    try:
        return EXTENSION_FORMAT_MAP[ext]
    except KeyError:
        raise ValueError("can't guess RDF format for %s, please specify it using "
                         "`rdf_format`" % filename)


class RDFRegistryError(Exception):
    pass


class RDFRegistry(object):
    """Class inspired from yams.xy / cubes.dataio.xy and holding static information about how to
    convert from a Yams model to RDF and the other way around.
    """
    def __init__(self):
        self.prefixes = {}
        self.etype2rdf = {}
        self.attr2rdf = {}
        self.rel2rdf = {}

    def normalize_uri(self, uri, raise_on_error=False):
        return normalize_uri(uri, self.prefixes, raise_on_error=raise_on_error)

    def register_prefix(self, prefix, namespace, overwrite=False):
        """Associate a prefix to a namespace. If the prefix is already registered to a different
        namespace, an exception is raised unless overwrite is True. Registered prefixes may be used
        in RDF snippets used in `register_*` methods.
        """
        if not overwrite and self.prefixes.get(prefix, namespace) != namespace:
            raise RDFRegistryError('prefix %r is already defined with different value %r'
                                   % (prefix, self.prefixes[prefix]))
        self.prefixes[prefix] = namespace

    def register_etype_equivalence(self, etype, rdftype, overwrite=False):
        """Associate a Yams entity type to a RDF type. If the entity type is already registered to a
        different RDF type, an exception is raised unless overwrite is True.
        """
        rdftype = self.normalize_uri(rdftype, raise_on_error=True)
        if not overwrite and self.etype2rdf.get(etype, rdftype) != rdftype:
            raise RDFRegistryError('entity type %r is already associated to RDF type %r'
                                   % (etype, self.etype2rdf[etype]))
        self.etype2rdf[etype] = rdftype

    def register_attribute_equivalence(self, etype, attr, rdftype, overwrite=False):
        """Associate a Yams entity attribute to a RDF predicate. If the entity attribute is already
        registered to a different RDF type, an exception is raised unless overwrite is True.
        """
        rdftype = self.normalize_uri(rdftype, raise_on_error=True)
        if not overwrite and self.attr2rdf.get((etype, attr), rdftype) != rdftype:
            raise RDFRegistryError('entity attribute %s.%r is already associated to RDF type %r'
                                   % (etype, attr, self.attr2rdf[(etype, attr)]))
        self.attr2rdf[(etype, attr)] = rdftype

    def register_relation_equivalence(self, subject_etype, rel, object_etype, rdftype,
                                      reverse=False):
        """Associate a Yams entity relation to a RDF predicated. The `reverse` flag may be used to
        indicate that in RDF, the subject and object are reversed (e.g. 'E1 yams_relation E2' is
        'E2 predicate E1 in RDF').
        """
        rdftype = self.normalize_uri(rdftype, raise_on_error=True)
        self.rel2rdf.setdefault((subject_etype, rel, object_etype), set()).add((rdftype, reverse))

    def predicates_for_subject_etype(self, etype):
        """Given a yams entity type, return (yams relation type, rdf predicate uri, reverse) 3-uple
        where the entity type is subject of the yams relation. `reverse` is a boolean flag telling
        whether the relation should be expected in the opposite direction in RDF (i.e. corresponding
        entity is the object of the rdf predicate), as they are not necessarily in the same order.
        """
        for (subject_etype, attr), rdftype in self.attr2rdf.items():
            if subject_etype == etype:
                yield attr, rdftype, False
        for (subject_etype, rel, _), rdf_relations in self.rel2rdf.items():
            if subject_etype == etype:
                for rdftype, reverse in rdf_relations:
                    yield rel, rdftype, reverse

    def predicates_for_object_etype(self, etype):
        """Given a yams entity type, return (yams relation type, rdf predicate uri, reverse) 3-uple
        where the entity type is object of the yams relation. `reverse` is a boolean flag telling
        wether the relation should be expected in the opposite direction in RDF (i.e. corresponding
        entity is the object of the rdf predicate), as they are not necessarily in the same order.
        """
        for (_, rel, object_etype), rdf_relations in self.rel2rdf.items():
            if object_etype == etype:
                for rdftype, reverse in rdf_relations:
                    yield rel, rdftype, reverse

    def additional_object_predicates(self, etype):
        """Given a yams entity type, return (yams relation type, rdf predicate uri, reverse) 3-uple
        where the entity type is object of the yams relation. `reverse` is a boolean flag telling
        wether the relation should be expected in the opposite direction in RDF (i.e. corresponding
        entity is the object of the rdf predicate), as they are not necessarily in the same order.

        Only relations whose subject is not a registered entity types (i.e. have had a mapping
        registered by :meth:`register_etype_equivalence`) will be returned.
        """
        for (subject_etype, rel, object_etype), rdf_relations in self.rel2rdf.items():
            if object_etype == etype and subject_etype not in self.etype2rdf:
                for rdftype, reverse in rdf_relations:
                    yield rel, rdftype, reverse


class AbstractRDFGraph(object):
    """Abstract base class for RDF graph implementation."""

    _backend_format_map = {}

    @property
    def supported_rdf_formats(self):
        """Return the set of RDF formats supported by the backend."""
        return set(self._backend_format_map)

    @property
    def uri(self):
        """Return the class used to represent URI for the concrete backend."""
        return self._uri_cls

    def load(self, source, rdf_format=None):
        """Add RDF triplets from a file, file path or URL `source` into the graph.

        Allowed values for optional `rdf_format` argument depends on the concret implementation,
        whose supported formats are listed in the `supported_rdf_formats` attribute. If not
        specified, it will be guessed from the file or URL extension.

        Raises `ValueError` if format is unknown or unspecified but can't be guessed.
        """
        if rdf_format is None:
            fname = getattr(source, 'name', source)  # get filename if source is a file.
            rdf_format = guess_rdf_format(fname)
        if rdf_format not in self.supported_rdf_formats:
            raise ValueError('RDF format %r is not supported' % rdf_format)
        self._load(source, self._backend_format_map[rdf_format])

    def _load(self, source, rdf_format):
        raise NotImplementedError

    def dump(self, fobj, rdf_format='xml'):
        """Dump the graph into the given file like object, using `rdf_format` representation.
        Allowed values for the optional `rdf_format` depends on the concret implementation, whose
        supported formats are listed in the `supported_rdf_formats` attribute. If not specified,
        RDF/XML will be used.

        Raises `ValueError` if format is unknown.
        """
        if rdf_format not in self.supported_rdf_formats:
            raise ValueError('RDF format %r is not supported' % rdf_format)
        self._dump(fobj, self._backend_format_map[rdf_format])

    def _dump(self, fobj, rdf_format):
        raise NotImplementedError

    def add(self, subj, predicate, obj):
        """Add statement to graph. subject and predicate are expected to be URIs and object may
        be either a URI or a literal value.
        """
        assert isinstance(subj, self.uri)
        assert isinstance(predicate, self.uri)
        if not isinstance(obj, self.uri):
            if isinstance(obj, unicode_with_language):
                try:
                    obj = self._literal(text_type(obj), lang=obj.lang)
                # an exception may be raised if the library doesn't like specified
                # language (e.g. rdflib check them using a regexp), but we don't
                # know at this point which kind of exception so catch Exception
                except Exception:
                    obj = self._literal(obj)
            else:
                obj = self._literal(obj)
        self._add(subj, predicate, obj)

    def _add(self, subj, predicate, obj):
        raise NotImplementedError


class LibRDFRDFGraph(AbstractRDFGraph):
    """redland's librdf based RDF graph"""
    _backend_format_map = {'nt': 'ntriples',
                           'n3': 'turtle',
                           'xml': 'rdfxml',
                           }
    _py_xsd_map = {
        int: 'http://www.w3.org/2001/XMLSchema#integer',
    }
    if PY2:
        _py_xsd_map[long] = 'http://www.w3.org/2001/XMLSchema#long'  # noqa

    _xsd_py_map = dict((v, k) for k, v in _py_xsd_map.items())

    def __init__(self, options_string="new='yes',hash-type='memory',dir='.'"):
        import RDF
        self._uri_cls = RDF.Uri
        storage = RDF.HashStorage("test", options_string)
        self._model = RDF.Model(storage)
        self._parser = RDF.Parser
        self._serializer = RDF.Serializer
        self._stmt = RDF.Statement
        node = RDF.Node
        self._node = node
        self._uri_node = lambda x: node(uri=x) if isinstance(x, RDF.Uri) else node(uri_string=x)

    def _load(self, source, rdf_format):
        """Add RDF triplets from a file path or URL into the graph."""
        assert isinstance(source, string_types), \
            '{0}._load() expect a path or an URL as source'.format(self.__class__.__name__)
        if ':/' not in source:
            source = 'file://' + abspath(source)
        parser = self._parser(name=rdf_format)
        parser.parse_into_model(self._model, self.uri(string=source))
        self._model.sync()  # in case model use a persistent storage

    def _dump(self, fobj, rdf_format):
        """Dump the graph into the given file like object, using `rdf_format` representation.
        """
        serializer = self._serializer(name=rdf_format)
        fobj.write(serializer.serialize_model_to_string(self._model))

    def _add(self, subj, predicate, obj):
        """Add RDF triplet to the graph."""
        self._model.add_statement(self._stmt(subj, predicate, obj))

    def _literal(self, value, lang=None):
        """Given a typed python value return proper RDFLib node."""
        if lang is None:
            try:
                datatype = self._py_xsd_map[type(value)]
            except KeyError:
                return self._node(literal=text_type(value))
            else:
                datatype = self.uri(datatype)
                return self._node(literal=text_type(value), datatype=datatype)
        return self._node(literal=value, language=str(lang))

    def _py_literal(self, node):
        """Given a RDFLib node return a typed python value."""
        try:
            string_value, language, datatype_uri = node.literal
        except AttributeError:
            # old librdf/redland version (e.g. centos6 has 1.0.7.1)
            string_value = node.literal_value['string']
            language = node.literal_value['language']
            datatype_uri = node.literal_value['datatype']
        if language is not None:
            return unicode_with_language(string_value, language)
        if datatype_uri is not None:
            datatype_uri = str(datatype_uri)
            if datatype_uri in self._xsd_py_map:
                return self._xsd_py_map[datatype_uri](string_value)
        return string_value

    def uris_for_type(self, type_uri):
        """Yield URIs of the given RDF type"""
        qs = self._stmt(subject=None,
                        predicate=self._uri_node(TYPE_PREDICATE_URI),
                        object=self._uri_node(type_uri))
        for statement in self._model.find_statements(qs):
            yield str(statement.subject.uri)

    def triples(self):
        """Yield every triples in the graph."""
        qs = self._stmt(subject=None, predicate=None, object=None)
        for statement in self._model.find_statements(qs):
            subject = str(statement.subject.uri)
            predicate = str(statement.predicate.uri)
            object = statement.object
            if object.is_literal():
                object = self._py_literal(statement.object)
            elif object.is_blank():
                object = str(statement.object.blank_identifier)
            else:
                object = str(statement.object.uri)
            yield subject, predicate, object

    def objects(self, entity_uri, predicate_uri):
        """Yield object URIs or literals that are linked to `entity_uri` through
        `predicate_uri`.
        """
        qs = self._stmt(subject=self._uri_node(entity_uri),
                        predicate=self._uri_node(predicate_uri),
                        object=None)
        for statement in self._model.find_statements(qs):
            if statement.object.is_literal():
                yield self._py_literal(statement.object)
            elif statement.object.is_blank():
                yield str(statement.object.blank_identifier)
            else:
                yield str(statement.object.uri)

    def subjects(self, predicate_uri, entity_uri):
        """Yield subject URIs that are linked to `entity_uri` through
        `predicate_uri`.
        """
        qs = self._stmt(subject=None,
                        predicate=self._uri_node(predicate_uri),
                        object=self._uri_node(entity_uri))
        for statement in self._model.find_statements(qs):
            yield str(statement.subject.uri)


class RDFLibRDFGraph(AbstractRDFGraph):
    """rdflib based RDF graph (http://rdflib.readthedocs.org)"""
    _backend_format_map = dict((x, x) for x in ('xml', 'n3', 'nt', 'rdfa'))
    _backend_format_map['pretty-xml'] = 'xml'

    def __init__(self):
        import rdflib
        self._uri_cls = rdflib.URIRef
        self._namespace = rdflib.namespace
        self._literal = rdflib.Literal
        self._graph = rdflib.ConjunctiveGraph()
        from rdflib.plugin import register, Parser
        register('text/rdf+n3', Parser, 'rdflib.plugins.parsers.notation3', 'N3Parser')

    def _load(self, source, rdf_format=None):
        """Add RDF triplets from a file stream, path or URL into the graph. `rdf_format` may be one
        of 'nt', 'n3' or 'xml'. If not specified, it will be guessed from the file or URL extension.
        """
        self._graph.parse(source, format=rdf_format)

    def _dump(self, fobj, rdf_format):
        """Dump the graph into the given file like object, using `rdf_format` representation.
        """
        self._graph.serialize(fobj, format=rdf_format)

    def _add(self, subj, predicate, obj):
        """Add RDF triplet to the graph."""
        self._graph.add((subj, predicate, obj))

    def uris_for_type(self, type_uri):
        """Return an iterator on URIs of the given RDF type"""
        for subj in self._graph.subjects(self._namespace.RDF.type, self.uri(type_uri)):
            yield text_type(subj)

    def triples(self):
        """Yield every triples in the graph."""
        for subj, pred, obj in self._graph.triples((None, None, None)):
            if isinstance(obj, self.uri):
                obj = text_type(obj)
            elif getattr(obj, 'language', None) is not None:
                obj = unicode_with_language(obj.toPython(), obj.language)
            else:
                obj = obj.toPython()
            yield text_type(subj), text_type(pred), obj

    def objects(self, entity_uri, predicate_uri):
        """Return an iterator on object URIs or literals that are linked to `entity_uri` through
        `predicate_uri`.
        """
        for obj in self._graph.objects(self.uri(entity_uri), self.uri(predicate_uri)):
            if isinstance(obj, self.uri):
                yield text_type(obj)
            elif getattr(obj, 'language', None) is not None:
                yield unicode_with_language(obj.toPython(), obj.language)
            else:
                yield obj.toPython()

    def subjects(self, predicate_uri, entity_uri):
        """Return an iterator on subject URIs that are linked to `entity_uri` through
        `predicate_uri`.
        """
        for subj in self._graph.subjects(self.uri(predicate_uri), self.uri(entity_uri)):
            yield text_type(subj)


def default_graph():
    """Return a default graph instance depending on installed software."""
    try:
        return RDFLibRDFGraph()
    except ImportError:
        try:
            return LibRDFRDFGraph()
        except ImportError:
            raise ValueError('neither rdflib nor RDF library installed')


def rdf_graph_to_entities(reg, graph, etypes, output_cls=ExtEntity):
    """Turns RDF data into an transitional ExtEntity representation that may
    be then imported into some database easily.

    Mapping is done using a RDF registry `reg` (see :class:`RDFRegistry`).
    """
    for etype in etypes:
        type_uri = reg.etype2rdf[etype]
        for uri in graph.uris_for_type(type_uri):
            extentity = output_cls(etype, uri)
            for rtype, predicate_uri, reverse in reg.predicates_for_subject_etype(etype):
                if reverse:
                    uris = graph.subjects(predicate_uri, uri)
                else:
                    uris = graph.objects(uri, predicate_uri)
                uris = set(uris)  # may be a generator
                if uris:
                    extentity.values[rtype] = uris
            for rtype, predicate_uri, reverse in reg.additional_object_predicates(etype):
                if reverse:
                    uris = graph.objects(uri, predicate_uri)
                else:
                    uris = graph.subjects(predicate_uri, uri)
                uris = set(uris)  # may be a generator
                if uris:
                    extentity.values[rtype] = uris
            yield extentity


def entities_stack(entities):
    """Coroutine function responsible for yielding entities from a stack of unique
    values that can be updated by calling the `send()` method.
    """
    entities = set(entities)
    processed = set()
    while entities:
        current = entities.pop()
        newones = yield current
        processed.add(current.eid)
        if newones:
            # Update the "stack" with new inputs.
            entities.update(x for x in newones if x.eid not in processed)


class RDFGraphGenerator(object):
    """RDF Graph generator using an RDF graph implementation.

    Fill the graph by calling `add_entity` with each entity to add to the graph.
    """
    def __init__(self, graph):
        self._graph = graph
        self._type_predicate = graph.uri(TYPE_PREDICATE_URI)
        self._same_as_predicate = graph.uri(SAME_AS_PREDICATE_URI)

    @classmethod
    def same_as_uris(cls, entity):
        """Return URIs of the given entity that should be linked using the `same_as` relation."""
        if urlparse(cls.canonical_uri(entity)).netloc != urlparse(entity.cwuri).netloc:
            yield entity.cwuri

    @staticmethod
    def canonical_uri(entity):
        """Return URIs of the given entity that should be linked using the `same_as` relation."""
        return permanent_url(entity)

    def add_entity(self, entity, reg):
        """Add information about a single entity as defined in the given RDF registry into the RDF
        graph and return related entities for eventual further processing.
        """
        graph = self._graph
        etype = entity.cw_etype
        if etype not in reg.etype2rdf:
            return ()
        uri = graph.uri(self.canonical_uri(entity))
        graph.add(uri, self._type_predicate, graph.uri(reg.etype2rdf[etype]))
        for same_as_uri in self.same_as_uris(entity):
            graph.add(uri, self._same_as_predicate, graph.uri(same_as_uri))
        related = set()
        for rtype, predicate_uri, reverse in reg.predicates_for_subject_etype(etype):
            if callable(rtype):
                values = rtype(entity)
            else:
                values = getattr(entity, rtype)
            if values is None:
                continue
            if isinstance(values, (tuple, list)):  # relation
                self._add_relations(uri, predicate_uri, reverse, values)
                related.update(values)
            else:  # attribute.
                assert not reverse, (uri, rtype, predicate_uri, reverse)
                graph.add(uri, graph.uri(predicate_uri), values)
        for rtype, predicate_uri, reverse in reg.predicates_for_object_etype(etype):
            try:
                values = getattr(entity, 'reverse_' + rtype)
            except AttributeError:
                # symmetric relations have no 'reverse_' attribute
                # XXX what if this is simply a bad mapping?
                continue
            related.update(values)
            self._add_relations(uri, predicate_uri, not reverse, values)
        return related

    def _add_relations(self, uri, predicate_uri, reverse, related_entities):
        """Add information about relations with `predicate_uri` between entity with `uri` and
        `related_entities` to the graph.
        """
        graph = self._graph
        predicate = graph.uri(predicate_uri)
        for related_entity in related_entities:
            r_uri = graph.uri(self.canonical_uri(related_entity))
            if reverse:
                graph.add(r_uri, predicate, uri)
            else:
                graph.add(uri, predicate, r_uri)



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from logilab.common.testlib import TestCase

import cubicweb.devtools  # noqa set $PYTHONPATH

from cubes.skos import lcsv
from cubes.skos.rdfio import unicode_with_language as ul


class LCSV2RDFTC(TestCase):

    def test_missing_prolog_column(self):
        stream = open(self.datapath('lcsv_example_missing_prolog.csv'))
        with self.assertRaises(lcsv.InvalidLCSVFile) as cm:
            lcsv.LCSV2RDF(stream, '\t', 'utf-8')
        self.assertIn("missing prolog column", str(cm.exception))

    def test_missing_id_column(self):
        stream = open(self.datapath('lcsv_example_missing_id.csv'))
        with self.assertRaises(lcsv.InvalidLCSVFile) as cm:
            lcsv.LCSV2RDF(stream, '\t', 'utf-8')
        self.assertIn("missing $id column", str(cm.exception))

    def test_lcsv_parsing(self):
        fpath = self.datapath('lcsv_example_shortened.csv')
        lcsv2rdf = lcsv.LCSV2RDF(open(fpath), '\t', 'utf-8',
                                 uri_generator=lambda x: x, default_lang='es')
        self.assertEqual(set(list(lcsv2rdf.triples())),
                         set([
                             ('#1', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
                              'http://www.w3.org/2004/02/skos/core#Concept'),
                             ('#1', 'http://www.w3.org/2004/02/skos/core#definition',
                              ul(u"Définition de l'organisation politique de l'organisme,", 'fr')),
                             ('#1', 'http://www.w3.org/2004/02/skos/core#prefLabel',
                              ul(u"Vie politique", 'es')),
                             ('#2', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
                              'http://www.w3.org/2004/02/skos/core#Concept'),
                             ('#2', 'http://www.w3.org/2004/02/skos/core#definition',
                              ul(u"Définition (évolution) des règles de fonctionnement", 'fr')),
                             ('#2', 'http://www.w3.org/2004/02/skos/core#prefLabel',
                              ul(u"Assemblée délibérante", 'es')),
                             ('#2', 'http://www.w3.org/2004/02/skos/core#broader',
                              '#1'),
                             ('#3', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
                              'http://www.w3.org/2004/02/skos/core#Concept'),
                             ('#3', 'http://www.w3.org/2004/02/skos/core#definition',
                              ul(u"Création volontaire ou en application de la loi", 'fr')),
                             ('#3', 'http://www.w3.org/2004/02/skos/core#prefLabel',
                              ul(u"Instances consultatives", 'es')),
                             ('#3', 'http://www.w3.org/2004/02/skos/core#broader',
                              '#1'),
                             ('#4', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
                              'http://www.w3.org/2004/02/skos/core#Concept'),
                             ('#4', 'http://www.w3.org/2004/02/skos/core#definition',
                              ul(u"Fonction de définition d'objectifs à long terme", 'fr')),
                             ('#4', 'http://www.w3.org/2004/02/skos/core#prefLabel',
                              ul(u"Pilotage de l'organisation", 'es')),
                             ('#5', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
                              'http://www.w3.org/2004/02/skos/core#Concept'),
                             ('#5', 'http://www.w3.org/2004/02/skos/core#definition',
                              ul(u"Définition du projet d'administration", 'fr')),
                             ('#5', 'http://www.w3.org/2004/02/skos/core#prefLabel',
                              ul(u"Pilotage de Bordeaux Metropole", 'es')),
                             ('#5', 'http://www.w3.org/2004/02/skos/core#broader',
                              '#4')
                         ]))

    def test_lcsv_parsing_sniff(self):
        fpath = self.datapath('lcsv_example_shortened.csv')
        lcsv2rdf = lcsv.LCSV2RDF(open(fpath),
                                 uri_generator=lambda x: x, default_lang='es')
        self.assertEqual(len(list(lcsv2rdf.triples())), 18)


if __name__ == "__main__":
    from unittest import main
    main()
